/*
	Visual Sciences .VSR extraction tool
	by Revenant - 25 Aug 2011
	
	Not fully implemented, but does enough to fulfill its intended purpose.
	May or may not work correctly in *nix.
	
	d@revenant1.net
*/
#include <stdio.h>
#include <string.h>
#include <time.h>
#include <stdlib.h>

#ifdef _WIN32
#include <direct.h>
#else
#include <sys/stat.h>
#define _mkdir(x) (mkdir(x, 666))
#endif

typedef struct {
	char type[5];
	int size;
	
	//CRID chunk info: name pointer, unknown int, data pointer
	struct {
		int name_offset;
		int unknown;
		int data_offset;
	} dir_data;
} chunk_t;

//size of directory entry in bytes
#define DIR_SIZE 36

typedef struct {
	int name_pointer;
	int file_num;
	char target_type[4];
	int chunk_pointer;
	
	//todo: structs for remaining info on each chunk type
} dir_t;

int totalfiles;

int walk(FILE *fp, int offset, const char *outdir);
int save(FILE *fp, const dir_t *file, const char *outdir, const char *path);

/*
	Traverse a directory in the archive, saving all files to disk.
	Recursively traverses all subdirectories as well.
*/
int walk(FILE *fp, int offset, const char *outdir) {
	chunk_t dir;
	dir_t file;
	
	char path[256];
	char filename[32];
	char filetype[6];
	int lastnum;
	int filesread;
		
	//seek to directory's offset in file
	if (fseek(fp, offset, SEEK_SET)) {
		printf("Error seeking to directory start (offset %x)\n", offset);
		return -1;
	}
	
	//read generic chunk info
	fread(dir.type, 4, 1, fp);
	dir.type[4] = '\0';
	fread(&dir.size, 4, 1, fp);
	
	//verify chunk type
	if (memcmp(dir.type, "CRID", 4)) {
		printf("error: expected \"CRID\", got \"%s\"\n", dir.type);
		return -1;
	}
	
	//read directory info
	fread(&dir.dir_data, sizeof(dir.dir_data), 1, fp);
	
	//printf("walking directory %s\t(%d bytes large)\n", outdir, dir.size);
	
	//create output directory
	_mkdir(outdir);
	
	//seek to directory entries and print each file in directory
	if (fseek(fp, dir.dir_data.data_offset, SEEK_SET)) {
		printf("Error seeking to directory entries (offset %x)\n", dir.dir_data.data_offset);
		return -1;
	}
	
	//read first file info
	filesread = 0;
	lastnum = -1;
	
	//keep reading directory entries until the entry's file number is inconsistent with the
	//previous file's number (since i don't see an obvious way to just get the number of files)
	while (1) {
		//read the file corresponding to the current directory entry
		if (fseek(fp, dir.dir_data.data_offset + DIR_SIZE * filesread, SEEK_SET)) {
			printf("Error seeking to directory entry #%d (offset %x)\n", filesread, dir.dir_data.data_offset);
			return -1;
		}
		fread(&file, sizeof(dir_t), 1, fp);
		
		//if filenum isn't 1 more than the previous file's number, then this isn't a valid file
		//meaning we've reached the end of the directory and are finished.
		//(assumes there is always at least one file in a directory)
		if (lastnum > -1 && file.file_num != lastnum + 1)
			break;
		
		//printf("file number: %d -> %d\n", lastnum, file.file_num);
		lastnum = file.file_num;
	
		//read file name
		if (fseek(fp, file.name_pointer, SEEK_SET)) {
			printf("Error seeking to filename (offset %x)\n", dir.dir_data.data_offset);
			return -1;
		}
		
		fread(filename, 32, 1, fp);
		
		//printf("found [%s] %s\n", file.target_type, filename);
		
		strcpy(path, outdir);
		strcat(path, "/");
		strcat(path, filename);
		
		//Check type of target chunk - if it's a directory, walk it
		//if it's a regular file, save it
		int result;
		if (!memcmp(file.target_type, "CRID", 4)) {
			result = walk(fp, file.chunk_pointer, path);
		} else {
			//reverse chunk ID to use as file extension
			filetype[0] = '.';
			for (int i = 0; i < 4; i++) {
				char c = file.target_type[i];
				if (c == ' ')
					filetype[4-i] = '\0';
				else filetype[4-i] = c;
			}
			filetype[5] = '\0';
			strcat(path, filetype);
			
			result = save(fp, &file, outdir, path);
		}
		
		//if there was an error saving the file, then abort 
		if (result) return result;
		
		totalfiles++;
		filesread++;
	} 
	
	return 0;
}

/*
	Write a chunk to disk based on its entry in a directory.
*/
int save(FILE *fp, const dir_t *file, const char *outdir, const char *path) {
	chunk_t chunk;
	void *bytes;
	FILE *outfile;

	//parse directory entry to read in chunk
	fseek(fp, file->chunk_pointer, SEEK_SET);
	
	//read generic chunk info
	fread(chunk.type, 4, 1, fp);
	chunk.type[4] = '\0';
	fread(&chunk.size, 4, 1, fp);
	
	bytes = malloc(chunk.size);
	fread(bytes, 1, chunk.size, fp);

	printf("writing \t%s\n",  path);
	
	//open output file
	outfile = fopen(path, "w+b");
	if (!outfile) {
		printf("Error opening %s\n", path);
		fclose(outfile);
		free(bytes);
		return -1;
	}
	
	//write file contents
	fseek(outfile, 0, SEEK_SET);
	
	if (!fwrite(bytes, 1, chunk.size, outfile)) {
		printf("Error writing to %s\n", path);
		fclose(outfile);
		free(bytes);
		return -1;
	}
	
	fclose(outfile);
	free(bytes);
	return 0;
}

int main(int argc, char **argv) {
	char *output_dir;
	FILE *infile;

	printf("Visual Sciences VSR archive extractor\n");

	if (argc < 2) {
		printf("Usage: unvsr infile [outdir]\n");
		printf("Reads archive \"infile\" and extracts files to the optional \"outdir\".\n");
		
		exit(-1);
	}
	
	//try to open input file
	infile = fopen(argv[1], "rb");
	if (!infile) {
		printf("Error opening %s\n", argv[1]);
		
		exit(-1);
	}
	
	//set output directory
	if (argc >= 3)
		output_dir = argv[2];
	else output_dir = ".";
	
	//begin traversing the archive, starting from the root directory
	totalfiles = 0;
	int result = walk(infile, 0, output_dir);
	printf("Processed %d files\\dirs in %.2f seconds.\n", totalfiles, (double)clock()/CLOCKS_PER_SEC);
	
	fclose(infile);
	
	return result;
}
